// Sprite.cpp

#include "StdAfx.h"

#include "Sprite.h"

// Usaremos esta estructura para guardar los diferentes sprites que tenemos en la textura
// Todo metido a mano, como los hombres XD
struct TextCoord
{
	float x, y;
	float width, height;
};

TextCoord boxArray[] = 
{
0.0f, 0.0f, 32.0f, 32.0f,
32.0f, 0.0f, 72.0f, 119.0f,
0.0f, 127.0f, 256.0f, 98.0f
};


Sprite::Sprite() : velY(0.0f)
{
	// Inicializamos el sprite
	int tile = rand() % 3;
	width = (int)boxArray[tile].width;
	height = (int)boxArray[tile].height;
	x = (float)(rand() % (640 - width));
	y = (float)(rand() % (480 / 2 - height));

	if(rand() % 2)
	{
		velX = (float)(rand() % 200);
	}
	else
	{
		velX = -(float)(rand() % 200);
	}
	
	angle = (float)(rand());

	// Creamos la matriz para transformas las coordenas de textura del quad
	// Esta matriz se usara en el VS para pillar el trozo de textura que deseemos
	// Primero se escala la "caja" de la textura, y luego se mueve sobre el trozo que queramos
	D3DXMATRIX matTrans;
	D3DXMatrixScaling(&matTexture, (float)width / 256.0f, (float)height / 256.0f, 0.0f);
	D3DXMatrixTranslation(&matTrans, (float)boxArray[tile].x / 256.0f, (float)boxArray[tile].y / 256.0f, 0.0f);
	D3DXMatrixMultiply(&matTexture, &matTexture, &matTrans);

	// Esto es MUY IMPORTANTE: los vertex shaders usan matrices transpuestas. Esto quiere decir que antes
	// de pasarle cualquier matriz a un VS hay que hacer el cambio. Por eso hacemos esta llamada
	D3DXMatrixTranspose(&matTexture, &matTexture);
}

void Sprite::draw(LPDIRECT3DDEVICE9 pD3DDevice, D3DXMATRIX *matViewProj)
{
	D3DXMATRIX matWorld;
	D3DXMATRIX matScaling;
	D3DXMATRIX matTranslation;
	D3DXMATRIX matRotation;

	// Aqui hacemos varias operaciones para conseguir el efecto de rotacion con el punto de referencia
	// en el centro del quad y moverlo por la pantalla rebotando contra el suelo.
	// El quad es de 1x1, asi que primero lo escalamos por width y height. Como usamos una proyeccion
	// ortogonal, tenemos una escala de 1:1, osea que una unida logica del mundo 3D es un pixel 
	// en el frame buffer.
	D3DXMatrixScaling(&matScaling, (float)width, (float)height, 0.0f);
	D3DXMatrixTranslation(&matTranslation, -(float)width / 2.0f, (float)height / 2.0f, 0.0f);
	D3DXMatrixMultiply(&matWorld, &matScaling, &matTranslation);
	D3DXMatrixRotationZ(&matRotation, angle);
	D3DXMatrixMultiply(&matWorld, &matWorld, &matRotation);
	D3DXMatrixTranslation(&matTranslation, x - SCREEN_WIDTH / 2.0f, (SCREEN_HEIGHT / 2.0f) - y, 0.0f);
	D3DXMatrixMultiply(&matWorld, &matWorld, &matTranslation);
	
	D3DXMatrixMultiply(&matWorld, &matWorld, matViewProj);
	// Ya tenemos la matriz de marras en matWorld. Ahora tenemos que "tumbarla" para que el VS
	// funcione correctamente. No os olvideis de esto, que sino os esperan interminables horas
	// de frustrante depuracion :P
	D3DXMatrixTranspose(&matWorld, &matWorld);
	
	// Le pasamos el matWorld al VS. Mirad los parametros en el SDK
	pD3DDevice->SetVertexShaderConstantF(0, (float *)&matWorld, 4);
	// Ahora le pasamos la matriz con la que se transformaran las coordenadas de textura
	pD3DDevice->SetVertexShaderConstantF(4, (float *)&matTexture, 4);

	D3DXVECTOR4 color(sinf(angle / 10), sinf(angle / 10), sinf(angle / 10), 1.0f - cosf(angle / 2));
	// Para terminar le pasamos el componente diffuse
	pD3DDevice->SetVertexShaderConstantF(8, (float *)&color, 1);
	
	// Dibujamos el sprite :D
	pD3DDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
}

void Sprite::update(float deltaTime)
{
	// Aqui esta el codigo que controla el sprite. Si teneis alguna duda o comentario ya sabeis
	angle += deltaTime * 5.0f;

	velY += 500.0f * deltaTime;
	
	if((x <= width / 2 && velX < 0) || (x + width / 2 >= 640 && velX > 0))
	{
		velX = -velX;
	}

	if(y + height / 2 >= 480 && velY > 0)
	{
		velY = -velY * 0.991f;
	}
	
	x += velX * deltaTime;
	y += velY * deltaTime;
}