//-----------------------------------------------------------------------------
// Ejemplo de Vertex Shader
// Creado en 2003 por AK47. Para cualquier duda o comentario: gammenon@msn.com
// Este ejemplo requiere el SDK de DirectX 9
//-----------------------------------------------------------------------------


#include "stdafx.h"
#include <mmsystem.h>
#include <d3dx9.h>
#include <d3dx9core.h>
#include <stdio.h>

#include "sprite.h"


// Formato del vertice
#define FVF_QUAD (D3DFVF_XYZ | D3DFVF_TEX1)

struct QUADVERTEX
{
	D3DXVECTOR3 vPos;
	D3DXVECTOR2 vTex;
};

struct QUAD
{
	QUADVERTEX V1;
	QUADVERTEX V2;
	QUADVERTEX V3;
	QUADVERTEX V4;
};

// Variables globales
LPDIRECT3D9 g_pD3D = NULL;
LPDIRECT3DDEVICE9 g_pd3dDevice = NULL;
LPDIRECT3DVERTEXBUFFER9 g_VBSprite = NULL; // El VB que usaran los sprites
LPD3DXFONT Font = NULL;
DWORD LastTime = 0;
int FPS = 0;
float TimeCounter = 0.0f;
int Frames = 0;
float g_Time = 0.0f;
LPDIRECT3DVERTEXSHADER9 g_pVertexShader = NULL; // El VS que usaremos
LPDIRECT3DVERTEXDECLARATION9 g_pVertexDeclaration = NULL; // La declaracion del VS
LPDIRECT3DTEXTURE9 g_Texture = NULL; // La textura que usaremos
D3DXMATRIX g_matViewProj; // La matriz de vista + proyeccion que le pasaremos al VS

const int spriteNumber = 10;
// El array de los sprites
Sprite *spriteArray = NULL;


void shutDown();


// En esta funcion inicializaremos todo lo referente a Direct3D
HRESULT InitD3D( HWND hWnd )
{
	// Creamos el objeto D3D
	g_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if(g_pD3D == NULL)
	{
		return E_FAIL;
	}

  // Inicializamos la estructura D3DPRESENT_PARAMETERS
  D3DPRESENT_PARAMETERS d3dpp; 
  ZeroMemory(&d3dpp, sizeof(d3dpp));
  d3dpp.Windowed = TRUE;
  d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	// Aqui podemos usar el flag D3DPRESENT_INTERVAL_IMMEDIATE, para que no espere al VSync
	// Por cierto, en mi maquina, usando D3DPRESENT_INTERVAL_DEFAULT da un rendimiento pobre o_O
	d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
  d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
  d3dpp.EnableAutoDepthStencil = TRUE;
  d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
	d3dpp.BackBufferWidth = SCREEN_WIDTH;
	d3dpp.BackBufferHeight = SCREEN_HEIGHT;
	d3dpp.hDeviceWindow = hWnd;
	d3dpp.BackBufferCount = 1;

  // Creamos el dispositivo
	// Usamos D3DCREATE_SOFTWARE_VERTEXPROCESSING para que el vertex shader se ejecutara por software
	// Asi funcionara en cualquier PC (presumiblemente...)
  if(FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                 D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                 &d3dpp, &g_pd3dDevice)))
  {
		return E_FAIL;
  }

	// Desactivamos el ZBuffer, la luz y la eliminacion de caras ocultas
  g_pd3dDevice->SetRenderState(D3DRS_ZENABLE, FALSE);
  g_pd3dDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
	g_pd3dDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);


	// Gestionamos lo referente a la fuente que usaremos para mostrar las estadisticas (FPS y tal)
	HFONT FontHandle;
	FontHandle = CreateFont(15, 0, 0, 0, 20, 0, 0, 0, 0, 0, 0, 0, 0, "System");
	D3DXCreateFont(g_pd3dDevice, FontHandle, &Font);

	srand (timeGetTime ());

	/*
	 Llegamos al meollo del asunto:
	 Definimos la declaracion del vertex shader. Con esto indicamos que es lo que le llegara al VS
	 mediante el stream de datos (que se fija con el conocido SetStreamSource)
	 La declaracion se compone de una serie de "lineas" que indican como esta organizado el stream.
	 En este ejemplo le pasaremos la posicion y las coordenadas de textura del vertice. La primera linea
	 indica la posicion, que son 3 componentes float, y le indicamos que son de la posicion mediante el
	 flag D3DDECLUSAGE_POSITION. El flag D3DDECLMETHOD_DEFAULT le indica al tesselator que no meta mano
	 en el asunto. Con la segunda linea le indicamos que le pasamos las coordenadas de textura, que son 2 floats y
	 que vienen 3 * sizeof(float) bytes despues de la posicion. El segundo parametro indica el offset
	 en el stream. Como se ve, en la primera linea, referente a la posicion, el offset es 0.
	 Con D3DDECL_END() le indicamos que hemos acabado la declaracion. Para saber el significado de cada
	 parametro y todo eso, consultad el SDK :)
	*/
	D3DVERTEXELEMENT9 decl[] = 
	{
		{ 0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
		{ 0, sizeof(float) * 3, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
		D3DDECL_END()
	};

	// Ahora creamos la declaracion
	HRESULT hr;
	if(FAILED(hr = g_pd3dDevice->CreateVertexDeclaration(decl, &g_pVertexDeclaration)))
  {
		return hr;
  }

	LPD3DXBUFFER pCode; // En este buffer se guardara el codigo del vertex shader
	LPD3DXBUFFER pErrorMsgs; // En este otro se guardara el mensaje de error en caso de algun fallo

	// Compilamos el vertex shader que se encuentra en el fichero "VertexShader.vsh"
	D3DXAssembleShaderFromFile("VertexShader.vsh", NULL, NULL, 0,  
															&pCode, &pErrorMsgs);

	// Ha habido un error...
	if(pErrorMsgs != NULL)
	{
		// Usad vuestro logger favorito ;)
		printf("Peaso error al crear el vertex buffer = %s", (char *)pErrorMsgs->GetBufferPointer());
		return E_FAIL;
	}
	// Todo perfecto, creamos el vertex shader
	g_pd3dDevice->CreateVertexShader((DWORD*)pCode->GetBufferPointer(),
																		&g_pVertexShader);

	// Ya tenemos el vertex shader listo para usar. Ahora borramos los buffers auxiliares
	pCode->Release();
	if(pErrorMsgs != NULL)
	{
		pErrorMsgs->Release();
	}

  return S_OK;
}

HRESULT InitGeometry()
{
	// Inicializamos el array de sprites
	spriteArray = new Sprite[spriteNumber];
	
	D3DXMATRIX matWorld;
	// Fijar la identidad como matriz de mundo
  g_pd3dDevice->SetTransform(D3DTS_WORLD, D3DXMatrixIdentity (&matWorld));

	// Creamos la matriz de vista
	D3DXMATRIX matView;
	D3DXVECTOR3 vEyePt(0.0f, 0.0f, -5.0f);
  D3DXVECTOR3 vLookatPt(0.0f, 0.0f, 0.0f);
  D3DXVECTOR3 vUpVec(0.0f, 1.0f, 0.0f);
  D3DXMatrixLookAtLH(&matView, &vEyePt, &vLookatPt, &vUpVec);
	// No es estrictamente necesario, pero bueno
  g_pd3dDevice->SetTransform(D3DTS_VIEW, &matView);


	// La matriz de proyeccion. Usaremos una matriz ortogonal
  D3DXMATRIX matProj;
	D3DXMatrixOrthoLH(&matProj, (float)SCREEN_WIDTH, (float)SCREEN_HEIGHT, 1.0f, 10.0f);
	// Tampoco es necesario, ya que las transformaciones las hara el VS
  g_pd3dDevice->SetTransform(D3DTS_PROJECTION, &matProj);

	// Ahora multiplicamos la matriz de vista y proyeccion para objentar la de vista-proyeccion, como
	// los mas listos habran podido adivinar. Esta matriz se la pasaremos al VS, ya que el VS se encarga
	// de todas las transformacion del vertice.
	D3DXMatrixMultiply(&g_matViewProj, &matView, &matProj);
	
	// Cargamos la textura con un colorkey de 0xFF00FF
	D3DXCreateTextureFromFileEx(g_pd3dDevice, "texture.bmp", 
		D3DX_DEFAULT, D3DX_DEFAULT, 0, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED,
		D3DX_FILTER_NONE, D3DX_FILTER_NONE, D3DCOLOR_ARGB(0x0, 0xFF, 0x0, 0xFF), NULL, NULL, &g_Texture);
	

	// Creamos el vertex buffer del sprite
	if (FAILED (g_pd3dDevice->CreateVertexBuffer (sizeof(QUAD), D3DUSAGE_WRITEONLY, 
		FVF_QUAD, D3DPOOL_MANAGED, &g_VBSprite, NULL)))
	{
		return E_FAIL;
	};

	// Rellenamos el vertex buffer con ricos datos :) Creamos un quad de 1x1, con la esquina superior
	// izquierda en (0,0) La z nos la suda :P
	QUAD Quad;
	Quad.V1.vPos = D3DXVECTOR3(0.0f, -1.0f, 0.0f);
	Quad.V2.vPos = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	Quad.V3.vPos = D3DXVECTOR3(1.0f, -1.0f, 0.0f);
	Quad.V4.vPos = D3DXVECTOR3(1.0f, 0.0f, 0.0f);

	// Estas coordenadas de textura las transformara el VS para que pille el trozo de textura que toquen
	Quad.V1.vTex = D3DXVECTOR2(0.0f, 1.0f);
	Quad.V2.vTex = D3DXVECTOR2(0.0f, 0.0f);
	Quad.V3.vTex = D3DXVECTOR2(1.0f, 1.0f);
	Quad.V4.vTex = D3DXVECTOR2(1.0f, 0.0f);

	// Bloqueamos el VB y lo rellenamos
	void *pVB;
	g_VBSprite->Lock(0, 0, &pVB, 0);

	memcpy(pVB, &Quad, sizeof(QUAD));

	g_VBSprite->Unlock();


	// Para terminar, fijamos diversos parametros para que D3D haga lo que deseamos
	// Con estas 3 lineas le indicamos a D3D que coja la textura y el componente diffuse y los sume.
	// Asi conseguiremos efectos como cambios de color en el sprite. Se puede cambiar el D3DTOP_ADD
	// por otras operaciones como MODULATE. Segun las necesidades, como siempre :)
	g_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	g_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
	g_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_ADD);
	// Con estas otras lineas indicamos que se tenga en cuenta el alpha de la textura (para no pintar
	// los pixels transparentes) y tambien el alpha del diffuse para hacer efectos de semitransparencias
	g_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
	g_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
	g_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE);
	g_pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
	g_pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
	g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
	// Exigimos a Direct3D que suavice las texturas >:D
	g_pd3dDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	g_pd3dDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

	// Estas operaciones, en un caso real se tendrian que hacer por frame, pero como es un ejemplo pos na ;P
	g_pd3dDevice->SetTexture(0, g_Texture);
	g_pd3dDevice->SetStreamSource(0, g_VBSprite, 0, sizeof (QUADVERTEX));
	g_pd3dDevice->SetVertexDeclaration(g_pVertexDeclaration);
	g_pd3dDevice->SetVertexShader(g_pVertexShader);

	// Kitto
  return S_OK;
}

VOID Render()
{
	// Primero controlamos el tiempo
	DWORD CurrentTime = timeGetTime ();

	g_Time = (CurrentTime - LastTime) / 1000.0f;

	LastTime = CurrentTime;

	TimeCounter += g_Time;
	Frames ++;
	if (TimeCounter >= 1.0f)
	{
		TimeCounter = 0.0f;
		FPS = Frames;
		Frames = 0;
	}

	if (g_Time >= 0.1)
	{
		return;
	}

	// Limpiamos el frame buffer con un verde chillon
	g_pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,255,0), 1.0f, 0);
  
  if(SUCCEEDED(g_pd3dDevice->BeginScene()))
  {
		// Dibujamos los sprites
		int i;
		for(i = 0; i < spriteNumber; i ++)
		{
			spriteArray[i].update(g_Time);
			spriteArray[i].draw(g_pd3dDevice, &g_matViewProj);
		}

		// Escribimos los FPS
		char Str [512];
		RECT Rect;
		SetRect(&Rect, 1, 1, 0, 0);
		sprintf(Str, "FPS = %d", FPS);
		Font->DrawText(Str, strlen (Str), &Rect, DT_LEFT, D3DCOLOR_XRGB (255, 0, 0));

		g_pd3dDevice->EndScene();
  }

	// Y enseamos todo lo que hemos cocinado
  g_pd3dDevice->Present(NULL, NULL, NULL, NULL);
}


//-----------------------------------------------------------------------------
// A partir de aqui es codigo de gestion de win32
//-----------------------------------------------------------------------------

// La funcion que atendera a los mensajes... lo de siempre
LRESULT WINAPI msgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  switch(msg)
  {
		case WM_DESTROY:
			shutDown();
			PostQuitMessage(0);
			return 0;
		case WM_KEYDOWN:
			TCHAR pressedKey = (TCHAR) wParam;
			if(pressedKey == VK_ESCAPE)
			{
				shutDown();
				PostQuitMessage(0);
			}
			return 0;
  }

  return DefWindowProc(hWnd, msg, wParam, lParam);
}

// Todo el asunto de Win32...
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
  WNDCLASSEX wc = {sizeof(WNDCLASSEX), CS_CLASSDC, msgProc, 0L, 0L, 
                   GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                   "Ejemplo de Vertex Shader", NULL };

  RegisterClassEx(&wc);

  HWND hWnd = CreateWindow("Ejemplo de Vertex Shader", "Ejemplo de Vertex Shader por AK47", 
                            WS_OVERLAPPEDWINDOW, 100, 100, 300, 300,
                            GetDesktopWindow(), NULL, wc.hInstance, NULL);

	// Ajustamos la ventana para que el area cliente sea SCREEN_WIDTH * SCREEN_HEIGHT
	RECT tempRect;
	SetRect(&tempRect, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
	AdjustWindowRectEx(&tempRect,
										GetWindowLong (hWnd, GWL_STYLE),
										GetMenu (hWnd) != NULL,
										GetWindowLong (hWnd, GWL_EXSTYLE));

	SetWindowPos (hWnd, NULL, 0, 0, tempRect.right - tempRect.left, tempRect.bottom - tempRect.top, SWP_NOZORDER | SWP_NOACTIVATE);

  if(SUCCEEDED(InitD3D(hWnd)))
  { 
    if(SUCCEEDED(InitGeometry()))
    {
      ShowWindow(hWnd, SW_SHOWDEFAULT);
      UpdateWindow(hWnd);

      MSG msg; 
      ZeroMemory(&msg, sizeof(msg));
      while(msg.message!=WM_QUIT)
      {
        if(PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE))
        {
          TranslateMessage(&msg);
          DispatchMessage(&msg);
        }
        else
				{
					Render();
				}
      }
    }
  }

  UnregisterClass("Ejemplo de Vertex Shader", wc.hInstance);
  return 0;
}

void shutDown()
{
	// Liberamos la preciosa memoria reservada para los sprites
	delete [] spriteArray;
	// Y tambien en lo referente a Direct3D
	if(g_Texture != NULL)
	{
		g_Texture->Release();
	}
	if(g_pVertexDeclaration != NULL)
	{
		g_pVertexDeclaration->Release();
	}
	if(g_pVertexShader != NULL)
	{
		g_pVertexShader->Release();
	}
	if(Font != NULL)
	{
		Font->Release();
	}
	if(g_VBSprite != NULL)
	{
		g_VBSprite->Release();
	}
	if(g_pd3dDevice != NULL)
	{
		g_pd3dDevice->Release();
	}
	if(g_pD3D != NULL)
	{
		g_pD3D->Release();
	}
}